# gophian -- tools to help with Debianizing Go software
# Copyright (C) 2024-2025 Maytham Alsudany <maytha8thedev@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from typing import List, Optional

import click
import requests
from treelib import Tree

from gophian.__about__ import BUGS_URL
from gophian.error import ExecutionError
from gophian.packages import DebianGolangPackages
from gophian.session import Session
from gophian.vcs import package_from_import_path


@click.command()
@click.argument("importpath")
@click.option("--quiet/--no-quiet", default=False, help="Don't print warnings.")
@click.option(
    "--warn-packaged/--no-warn-packaged",
    default=True,
    help="Warn if the given Go package has already been packaged for Debian.",
)
@click.option(
    "--include-packaged-deps/--no-include-packaged-deps",
    default=True,
    help="Include packaged dependencies in the estimate.",
)
@click.option(
    "--include-version-conflicts/--no-include-version-conflicts",
    default=False,
    help=(
        "Include any version conflicts between the package's go.mod and the"
        "version in Debian unstable. Ignored if --no-include-packaged-deps is"
        "passed."
    ),
)
@click.option(
    "--max-depth",
    required=False,
    help="Maximum recursion depth of unpackaged dependencies.",
)
@click.pass_context
def estimate(
    context: click.Context,
    importpath: str,
    quiet: bool,
    warn_packaged: bool,
    include_packaged_deps: bool,
    include_version_conflicts: bool,
    max_depth: Optional[int],
) -> None:
    """
    Estimate work required to Debianize a given Go package.
    """

    requests_session: requests.Session = context.obj

    if not quiet:
        click.secho("gophian is experimental software!", bold=True, fg="yellow")
        click.secho("Please report any problems to:", fg="yellow")
        click.secho(BUGS_URL, fg="yellow")

    try:
        package, _ = package_from_import_path(requests_session, importpath)
    except Exception as error:
        click.echo(error)
        click.secho("Did you specify a Go package import path?", fg="yellow")
        context.exit(1)

    if package != importpath:
        click.echo(f"Continuing with repo root {package} instead of {importpath}")

    debian_packages = DebianGolangPackages(requests_session)

    if debian_packages._check_for_package(package, quiet) and warn_packaged:
        if not quiet:
            click.secho("To ignore, pass the --no-warn-packaged flag.", fg="cyan")
        return

    with Session(requests_session) as session:
        seen: List[str] = []
        errors: List[ExecutionError] = []

        tree = Tree()
        tree.create_node(package, package)
        dependency_tree(
            session,
            seen,
            errors,
            package,
            include_packaged_deps,
            include_version_conflicts,
            debian_packages,
            tree,
            max_depth,
        )
        click.echo(tree.show(stdout=False))

        for error in errors:
            click.echo(error)


def dependency_tree(
    session: Session,
    seen: List[str],
    errors: List[ExecutionError],
    package: str,
    include_packaged_deps: bool,
    include_version_conflicts: bool,
    debian_packages: DebianGolangPackages,
    tree: Tree,
    max_depth: Optional[int],
    level: int = 1,
):
    try:
        go_package = session.go_get(package)
        dep_packages = go_package.find_dependencies()
    except ExecutionError as error:
        errors.append(error)
        tree.create_node(
            click.style("Error getting dependencies", fg="red"),
            f"{package}_error",
            parent=package,
        )
        return
    for dep_package, _ in dep_packages:
        if (
            (result := debian_packages.library_is_packaged(dep_package))
            and (debian_package := result[0])
            and (debian_package_version := result[1])
        ):
            if include_packaged_deps:
                node_text = click.style(
                    f"{dep_package} ({debian_package})",
                    fg="green",
                )
                if include_version_conflicts:
                    [go_mod_major, go_mod_minor] = go_package.deps[
                        dep_package
                    ].vstring.split(".")[:2]
                    [debian_major, debian_minor] = debian_package_version.vstring.split(
                        "."
                    )[:2]
                    if (
                        (
                            go_package.deps[dep_package].commit
                            and go_package.deps[dep_package].vstring
                            != debian_package_version.vstring
                        )
                        or (go_mod_major != debian_major)
                        or (go_mod_major == 0 and go_mod_minor != debian_minor)
                    ):
                        node_text += click.style(
                            f" [{debian_package_version} ≠ {go_package.deps[dep_package]}]",
                            fg="yellow",
                        )
                tree.create_node(
                    node_text,
                    f"{package}_{dep_package}",
                    parent=package,
                )
        elif dep_package in seen:
            tree.create_node(
                f"{dep_package}",
                f"{package}_{dep_package}",
                parent=package,
            )
        elif max_depth is None or (level + 1) > max_depth:
            tree.create_node(
                dep_package,
                dep_package,
                parent=package,
            )
            seen.append(dep_package)
            dependency_tree(
                session,
                seen,
                errors,
                dep_package,
                include_packaged_deps,
                include_version_conflicts,
                debian_packages,
                tree,
                max_depth,
                level + 1,
            )
